import time
import urllib.parse

from botocore.exceptions import ClientError
from .common import APP_CGAP, APP_FOURFRONT, EnvName, UrlString
from .env_utils import (
    is_cgap_env, is_fourfront_env, get_standard_mirror_env,
    compute_prd_env_for_project, get_env_real_url,
)
from .misc_utils import PRINT


FOURSIGHT_URL = 'https://foursight.4dnucleome.org/'

# FF_MAGIC_CNAME corresponds to data.4dnucleome.org
_FF_MAGIC_CNAME = 'fourfront-green.us-east-1.elasticbeanstalk.com'
# CGAP_MAGIC_CNAME corresponds to cgap.hms.harvard.edu
_CGAP_MAGIC_CNAME = 'fourfront-cgap.9wzadzju3p.us-east-1.elasticbeanstalk.com'

# FF_GOLDEN_DB is the database behind data.4dnucleome.org (and shared by staging.4dnucleome.org)
_FF_GOLDEN_DB = 'fourfront-production.co3gwj7b7tpq.us-east-1.rds.amazonaws.com'
# CGAP_GOLDEN_DB is the database behind cgap.hms.harvard.edu
_CGAP_GOLDEN_DB = 'fourfront-cgap.co3gwj7b7tpq.us-east-1.rds.amazonaws.com'


def describe_beanstalk_environments(client, **kwargs):
    """
    Generic function for retrying client.describe_environments to avoid
    AWS throttling errors. Passes all given kwargs to describe_environments

    Args:
        client (botocore.client.ElasticBeanstalk): boto3 client

    Returns:
        dict: response from client.describe_environments

    Raises:
        Exception: if a non-ClientError exception is encountered during
            describe_environments or cannot complete within retry framework
    """
    env_info = kwargs.get('EnvironmentNames', kwargs.get('ApplicationName', 'Unknown environment'))
    for retry in [1, 1, 1, 1, 2, 2, 2, 4, 4, 6, 8, 10, 12, 14, 16, 18, 20]:
        try:
            res = client.describe_environments(**kwargs)
        except ClientError as e:
            PRINT('Client exception encountered while getting BS info for %s. Error: %s' % (env_info, str(e)))
            time.sleep(retry)
        except Exception as e:
            PRINT('Unhandled exception encountered while getting BS info for %s. Error: %s' % (env_info, str(e)))
            raise e
        else:
            return res
    raise Exception('Could not describe Beanstalk environments due ClientErrors, likely throttled connections.')


def beanstalk_info(env):
    """
    This doesn't actually talk to a beanstalk any more, but now returns a subset of the information
    that a beanstalk info query would have returned, specifically:

      {"EnvironmentName": <env>, "CNAME": <portal-url-hostname>}


    Args:
        env (str): ElasticBeanstalk environment name

    Returns:
        dict: Environments result from describe_beanstalk_environments
    """
    return get_env_info(env)


def get_env_info(env):
    """
    This returns some information about the given environment in a format similar to what beanstalk_info used to
    return. (If there are other elements of that function's return value that are needed, this is the function to
    extend.) The return value for now is:

    Args:
        env (str): ElasticBeanstalk environment name

    Returns:
        a dictionary of the form {"EnvironmentName": <env>, "CNAME": <portal-url-hostname>}
    """
    url = get_env_real_url(env)
    parsed_url = urllib.parse.urlparse(url)

    return {
        'EnvironmentName': env,
        'CNAME': parsed_url.hostname,
    }


def get_beanstalk_real_url(env: EnvName) -> UrlString:
    """
    Return the real url for the portal with given environment name.
    Name can be a special name (like 'cgap', 'data', 'staging'),
    or an actual environment.

    Note that because we are all-containerized now, this is now table-driven and does not actually do a beanstalk
    call. We left the name for now, to not break things, but you should consider it deprecated and
    use get_env_real_url instead.

    Args:
        env (str): ElasticBeanstalk environment name

    Returns:
        str: url of the ElasticBeanstalk environment
    """

    return get_env_real_url(env)


def compute_ff_prd_env():  # a.k.a. "whodaman" (its historical name, which has gone away)
    """Returns the name of the current Fourfront production environment."""
    return compute_prd_env_for_project(APP_FOURFRONT)


def compute_ff_stg_env():
    """Returns the name of the current Fourfront staging environment."""
    return get_standard_mirror_env(compute_ff_prd_env())


def compute_cgap_prd_env():
    """Returns the name of the current CGAP production environment."""
    return compute_prd_env_for_project(APP_CGAP)


def compute_cgap_stg_env():
    """Returns the name of the current CGAP staging environment, or None if there is none."""
    return get_standard_mirror_env(compute_cgap_prd_env())


def compute_prd_env_for_env(envname):
    """Given an environment, returns the name of the prod environment for its owning project."""
    if is_cgap_env(envname):
        return compute_cgap_prd_env()
    elif is_fourfront_env(envname):
        return compute_ff_prd_env()
    else:
        raise ValueError("Unknown environment: %s" % envname)


def compute_stg_env_for_env(envname):
    return get_standard_mirror_env(compute_prd_env_for_env(envname))



import contextlib

import chardet
import csv
import glob
import io
import json
import openpyxl
import os
import re
import subprocess
import yaml

from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.workbook.workbook import Workbook
from tempfile import TemporaryFile, TemporaryDirectory
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, Union
from .common import AnyJsonData, Regexp, JsonSchema
from .lang_utils import conjoined_list, disjoined_list, maybe_pluralize
from .misc_utils import ignored, pad_to, JsonLinesReader, remove_suffix, to_camel_case


Header = str
Headers = List[str]
TabbedHeaders = Dict[str, Headers]
ParsedHeader = List[Union[str, int]]
ParsedHeaders = List[ParsedHeader]
TabbedParsedHeaders = Dict[str, ParsedHeaders]
SheetCellValue = Union[int, float, str]
SheetRow = List[SheetCellValue]
CsvReader = type(csv.reader(TemporaryFile()))
SheetData = List[dict]
TabbedSheetData = Dict[str, SheetData]
TabbedJsonSchemas = Dict[str, JsonSchema]
CommentRow = object()


class LoadFailure(Exception):
    """
    In general, we'd prefer to load up the spreadsheet with clumsy data that can then be validated in detail,
    but some errors are so confusing or so problematic that we need to just fail the load right away.
    """
    pass


class LoadArgumentsError(LoadFailure):
    """
    Errors of this class represent situations where we can't get started because
    there's a problem with the given arguments.
    """
    pass


class LoadTableError(LoadFailure):
    """
    Errors of this class represent situations where we can't get started because
    there's a problem with some table's syntax, for example headers that don't make sense.
    """
    pass


def unwanted_kwargs(*, context, kwargs, context_plural=False, detailed=False):
    if kwargs:
        unwanted = [f"{argname}={value!r}" if detailed else argname
                    for argname, value in kwargs.items()
                    if value is not None]
        if unwanted:
            does_not = "don't" if context_plural else "doesn't"
            raise LoadArgumentsError(f"{context} {does_not} use"
                                     f" {maybe_pluralize(unwanted, 'keyword argument')} {conjoined_list(unwanted)}.")


def prefer_number(value: SheetCellValue, kind='number'):
    """
    Given a string, if the string has number syntax, returns the number it represents. Otherwise, returns its argument.
    (It follows from this that if given an int or a float, it just returns that argument.)

    Using a kind= argument (as in kind='int' or kind='float') can better restrict what kind of number a string
    is coerced to, but it has no effect when the argument is a number, even a number of the wrong kind.

    :param value: a string, int, or float
    :param kind: one of 'number' or 'integer'
    :returns: the argument coerced to a number of the appropriate kind, if possible, or else the argument literally
    """
    if isinstance(value, str):  # the given value might be an int or float, in which case just fall through
        if not value:
            return None
        value = value
        ch0 = value[0]
        if ch0 == '+' or ch0 == '-' or ch0.isdigit():
            if kind == 'integer':
                try:
                    return int(value)
                except Exception:
                    pass
            if kind == 'number':
                try:
                    return float(value)
                except Exception:
                    pass
        # If we couldn't parse it as an int or float, fall through to returning the original value
        pass
    # NOTE WELL:
    #   In the case where we already have a number, even if it's the wrong type, we just leave it as we got it.
    #   The job of this function is not to do type enforcement or correctness checking, but rather to adjust
    #   for the fact that spreadsheets and csv files often pass string data where they mean to pass numbers.
    #   If some human has already been thinking about what to pass in a JSON or other such setting,
    #   this function is not trying to be smart enough to second-guess that.
    return value


def expand_string_escape_sequences(text: str) -> str:
    """
    Expands string escape sequences in a commonly used way.
    A backslash followed by one of the following characters is expanded as indicated:
        r (return or CR) - ASCII 13 decimal, 15 octal, 0d hex
        n (newline or linefeed or LF) - ASCII 10 decimal, 12 octal, 0a hex
        t (tab) - ASCII 9 decimal, 11 octal, 9 hex
        f (formfeed or page) - ASCII 12 decimal, 14 octal, 0c hex
        \\ (backslash) - ASCII 92 decimal, 134 octal, 5c hex
    In all other situations, the backslash is left uninterpreted.
    """
    s = io.StringIO()
    escaping = False
    for ch in text:
        if escaping:
            if ch == 'r':
                s.write('\r')
            elif ch == 't':
                s.write('\t')
            elif ch == 'n':
                s.write('\n')
            elif ch == '\\':
                s.write('\\')
            else:
                # Rather than err, just leave other sequences as-is.
                s.write(f"\\{ch}")
            escaping = False
        elif ch == '\\':
            escaping = True
        else:
            s.write(ch)
    return s.getvalue()


def open_unicode_text_input_file_respecting_byte_order_mark(filename):
    """
    Opens a file for text input, respecting a byte-order mark (BOM).
    """
    with io.open(filename, 'rb') as fp:
        leading_bytes = fp.read(4 * 8)  # 4 bytes is all we need
        bom_info = chardet.detect(leading_bytes, should_rename_legacy=True)
        detected_encoding = bom_info and bom_info.get('encoding')  # tread lightly
    use_encoding = 'utf-8' if detected_encoding == 'ascii' else detected_encoding
    return io.open(filename, 'r', encoding=use_encoding)


class AbstractTableSetManager:
    """
    The TableSetManager is the spanning class of anything that wants to be able to load a table set,
    regardless of what it wants to load it from. To do this, it must support a load method
    that takes a filename and returns the file content in the form:
        {
            "Sheet1": [
                          {...representation of row1 as some kind of dict...},
                          {...representation of row2 as some kind of dict...}
                      ],
            "Sheet2": [...],
            ...,
        }
    It also needs some implementation of the .tab_names property.
    Note that at this level of abstraction, we take no position on what form of representation is used
    for the rows, as long as it is JSON data of some kind. It might be
         {"col1": "val1", "col2": "val2", ...}
    or it might be something more structured like
         {"something": "val1", {"something_else": ["val2"]}}
    Additionally, the values stored might be altered as well. In particular, the most likely alteration
    is to turn "123" to 123 or "" to None, though the specifics of whether and how such transformations
    happen is not constrained by this class.
    """

    ALLOWED_FILE_EXTENSIONS: List[str] = []

    def __init__(self, filename: str, prefer_number: Optional[bool] = None, **kwargs):
        if prefer_number:
            # It's OK to pass prefer_number=None (meaning take the default) and prefer_number=False,
            # since that requires no action, but if a class wants to manage such preferences,
            # as happens in FlattenedTableSetManager, it will have to do it itself.
            raise ValueError(f"This class {self.__class__.__name__} does not implement prefer_number={prefer_number!r}")
        self.filename: str = filename
        unwanted_kwargs(context=self.__class__.__name__, kwargs=kwargs)

    @classmethod
    def load(cls, filename: str, **kwargs) -> TabbedSheetData:
        """
        Reads a filename and returns a dictionary that maps sheet names to rows of dictionary data.
        For more information, see documentation of AbstractTableSetManager.
        """
        raise NotImplementedError(f".load(...) is not implemented for {cls.__name__}.")  # noQA

    @property
    def tab_names(self, order: Optional[List[str]] = None) -> List[str]:
        raise NotImplementedError(f".tab_names is not implemented for {self.__class__.__name__}..")  # noQA

    def load_content(self, sheet_order: Optional[List[str]] = None) -> Any:
        raise NotImplementedError(f".load_content() is not implemented for {self.__class__.__name__}.")  # noQA


class BasicTableSetManager(AbstractTableSetManager):
    """
    A BasicTableManager provides some structure that most kinds of parsers will need.
    In particular, everything will likely need some way of storing headers and some way of storing content
    of each sheet. Even a csv file, which doesn't have multiple tabs can be seen as the degenerate case
    of this where there's only one set of headers and only one block of content.
    """

    def __init__(self, filename: str, **kwargs):
        super().__init__(filename=filename, **kwargs)
        self.headers_by_tab_name: Dict[str, Headers] = {}
        self.content_by_tab_name: Dict[str, SheetData] = {}
        self.reader_agent: Any = self._get_reader_agent()

    def tab_headers(self, tab_name: str) -> Headers:
        return self.headers_by_tab_name[tab_name]

    def tab_content(self, tab_name: str) -> List[AnyJsonData]:
        return self.content_by_tab_name[tab_name]

    @classmethod
    def _create_tab_processor_state(cls, tab_name: str) -> Any:
        """
        This method provides for the possibility that some parsers will want auxiliary state,
        (such as parsed headers or a line count or a table of temporary names for objects to cross-link
        or some other such feature) that it carries with it as it moves from line to line parsing things.
        Subclasses might therefore want to make this do something more interesting.
        """
        ignored(tab_name)  # subclasses might need this, but we don't
        return None

    def _get_reader_agent(self) -> Any:
        """This function is responsible for opening the workbook and returning a workbook object."""
        raise NotImplementedError(f"._get_reader_agent() is not implemented for {self.__class__.__name__}.")  # noQA


class FlattenedTableSetManager(BasicTableSetManager):
    """
    This is the base class for all workbook-like data sources, i.e., that may need to apply semantic processing.
    Those may be:
    * Excel workbook readers (.xlsx)
    * Comma-separated file readers (.csv)
    * Tab-separarated file readers (.tsv in most of the world, but Microsoft stupidly calls this .txt, outright
      refusing to write a .tsv file, so many people seem to compromise and call this .tsv.txt)
    There are two levels to each of these: a class that is not semantically interpreted,
    and a class that is semantically interpreted as an "item".

    This is NOT a parent class of these kinds of files, which we always take literally as if semantic processing
    were already done (in part so that they can be used to test the results of other formats):
    * Json files
    * Yaml files
    * Inserts directories
    * JsonLines files
    """

    @classmethod
    def load(cls, filename: str, **kwargs) -> AnyJsonData:
        if cls.ALLOWED_FILE_EXTENSIONS:
            if not any(filename.lower().endswith(suffix) for suffix in cls.ALLOWED_FILE_EXTENSIONS):
                raise LoadArgumentsError(f"The TableSetManager subclass {cls.__name__} expects only"
                                         f" {disjoined_list(cls.ALLOWED_FILE_EXTENSIONS)} filenames: {filename}")

        table_set_manager: FlattenedTableSetManager = cls(filename=filename, **kwargs)
        return table_set_manager.load_content()

    def __init__(self, filename: str, prefer_number: Optional[bool] = None, **kwargs):
        if prefer_number is None:  # i.e., no initial value specified
            prefer_number = True
        self.prefer_number: bool = prefer_number
        super().__init__(filename=filename, **kwargs)

    def _raw_row_generator_for_tab_name(self, tab_name: str) -> Iterable[SheetRow]:
        """
        Given a tab_name and a state (returned by _sheet_loader_state), return a generator for a set of row values.
        """
        raise NotImplementedError(f"._rows_for_tab_name(...) is not implemented for {self.__class__.__name__}.")  # noQA

    def _process_row(self, tab_name: str, state: Any, row: List[SheetCellValue]) -> AnyJsonData:
        """
        This needs to take a state and whatever represents a row and
        must return a list of objects representing column values.
        What constitutes a processed up to the class, but other than that the result must be a JSON dictionary.
        """
        raise NotImplementedError(f"._process_row(...) is not implemented for {self.__class__.__name__}.")  # noQA

    def load_content(self, sheet_order: Optional[List[str]] = None) -> AnyJsonData:
        for tab_name in self.tab_names(sheet_order):
            sheet_content = []
            state = self._create_tab_processor_state(tab_name)
            for row_data in self._raw_row_generator_for_tab_name(tab_name):
                processed_row_data: AnyJsonData = self._process_row(tab_name, state, row_data)
                if not processed_row_data:
                    break
                if processed_row_data is CommentRow:
                    continue
                sheet_content.append(processed_row_data)
            self.content_by_tab_name[tab_name.replace(" ", "")] = sheet_content
        return self.content_by_tab_name

    def parse_cell_value(self, value: SheetCellValue, override_prefer_number: Optional[bool] = None) -> AnyJsonData:
        if override_prefer_number is None:
            override_prefer_number = self.prefer_number
        return prefer_number(value) if override_prefer_number else value


class TableSetManagerRegistry:

    def __init__(self):
        self.manager_table: Dict[str, Type[AbstractTableSetManager]] = {}
        self.regexp_mappings: List[Tuple[Regexp, Type[AbstractTableSetManager]]] = []

    def register(self, regexp: Optional[str] = None):
        def _wrapped_register(class_to_register: Type[AbstractTableSetManager]):
            if regexp:
                self.regexp_mappings.append((re.compile(regexp), class_to_register))
            for ext in class_to_register.ALLOWED_FILE_EXTENSIONS:
                existing = self.manager_table.get(ext)
                if existing:
                    raise Exception(f"Tried to define {class_to_register} to extension {ext},"
                                    f" but {existing} already claimed that.")
                self.manager_table[ext] = class_to_register
            return class_to_register
        return _wrapped_register

    register1 = register

    def manager_for_filename(self, filename: str) -> Type[AbstractTableSetManager]:
        base: str = os.path.basename(filename)
        suffix_parts = base.split('.')[1:]
        if suffix_parts:
            for i in range(0, len(suffix_parts)):
                suffix = f".{'.'.join(suffix_parts[i:])}"
                found: Optional[Type[AbstractTableSetManager]] = self.manager_table.get(suffix)
                if found:
                    return found
        else:
            special_case: Optional[Type[AbstractTableSetManager]] = self.manager_for_special_filename(filename)
            if special_case:
                return special_case
        raise LoadArgumentsError(f"Unknown file type: {filename}")

    def manager_for_special_filename(self, filename: str) -> Optional[Type[AbstractTableSetManager]]:
        for pattern, manager_class in self.regexp_mappings:
            if pattern.match(filename):
                return manager_class
        return None


TABLE_SET_MANAGER_REGISTRY = TableSetManagerRegistry()


@TABLE_SET_MANAGER_REGISTRY.register()
class XlsxManager(FlattenedTableSetManager):
    """
    This implements the mechanism to get a series of rows out of the sheets in an XLSX file.
    """
    ALLOWED_FILE_EXTENSIONS = ['.xlsx']
    TERMINATE_ON_EMPTY_ROW = True
    CONVERT_VALUES_TO_STRING = True

    @classmethod
    def _all_rows(cls, sheet: Worksheet):
        row_max = sheet.max_row
        for row in range(2, row_max + 1):
            yield row

    @classmethod
    def _all_cols(cls, sheet: Worksheet):
        col_max = sheet.max_column
        for col in range(1, col_max + 1):
            yield col

    def tab_names(self, order: Optional[List[str]] = None) -> List[str]:
        def ordered_sheet_names(sheet_names: List[str]) -> List[str]:
            if not order:
                return sheet_names
            ordered_sheet_names = []
            for item in order:
                for sheet_name in sheet_names:
                    if to_camel_case(item.replace(" ", "")) == to_camel_case(sheet_name.replace(" ", "")):
                        ordered_sheet_names.append(sheet_name)
            for sheet_name in sheet_names:
                if sheet_name not in ordered_sheet_names:
                    ordered_sheet_names.append(sheet_name)
            return ordered_sheet_names
        return ordered_sheet_names(self.reader_agent.sheetnames)

    def _get_reader_agent(self) -> Workbook:
        return openpyxl.load_workbook(self.filename)

    def _raw_row_generator_for_tab_name(self, tab_name: str) -> Iterable[SheetRow]:
        sheet = self.reader_agent[tab_name]
        return (self._get_raw_row_content_tuple(sheet, row)
                for row in self._all_rows(sheet))

    def _get_raw_row_content_tuple(self, sheet: Worksheet, row: int) -> SheetRow:
        # return [str(sheet.cell(row=row, column=col).value) for col in self._all_cols(sheet)]
        results = []
        for col in self._all_cols(sheet):
            value = sheet.cell(row=row, column=col).value
#           if value is not None and XlsxManager.CONVERT_VALUES_TO_STRING:
#               value = str(value).strip()
            if XlsxManager.CONVERT_VALUES_TO_STRING:
                value = str(value).strip() if value is not None else ""
            results.append(value)
        return results

    def _create_tab_processor_state(self, tab_name: str) -> Headers:
        sheet = self.reader_agent[tab_name]
        """
        headers: Headers = [str(sheet.cell(row=1, column=col).value)
                            for col in self._all_cols(sheet)]
        """
        headers = []
        for col in self._all_cols(sheet):
            cell = sheet.cell(row=1, column=col).value
            if cell is not None and XlsxManager.CONVERT_VALUES_TO_STRING:
                cell = str(cell).strip()
            headers.append(cell)
        self.headers_by_tab_name[sheet.title] = XlsxManager.remove_trailing_none_values(headers)
        return self.headers_by_tab_name[sheet.title]

    def _process_row(self, tab_name: str, headers: Headers, row_data: SheetRow) -> AnyJsonData:
        ignored(tab_name)
        if XlsxManager.is_terminating_row(row_data):
            return None
        if XlsxManager.is_comment_row(row_data):
            return CommentRow
        if len(headers) < len(row_data):
            row_data = row_data[:len(headers)]
        override_prefer_number = False if XlsxManager.CONVERT_VALUES_TO_STRING else None
        return {headers[i]: self.parse_cell_value(row_datum, override_prefer_number=override_prefer_number)
                for i, row_datum in enumerate(row_data)}

    @staticmethod
    def is_terminating_row(row: List[Optional[Any]]) -> bool:
        # TODO: This is may change; currently an all blank row signals the end of input.
        return all(cell is None or cell == "" for cell in row) and XlsxManager.TERMINATE_ON_EMPTY_ROW

    @staticmethod
    def is_comment_row(row: Tuple[Optional[Any]]) -> bool:
        # TODO: This will probably change; currently a row starting only with #, *, or ^ signals a comment.
        for cell in row:
            if cell is None:
                continue
            if (cell := str(cell)).startswith("#") or cell.startswith("*") or cell.startswith("^"):
                return True
        return False

    @staticmethod
    def remove_trailing_none_values(values: List[Any]) -> List[Any]:
        for index in range(len(values) - 1, -1, -1):
            if values[index] is not None:
                break
        else:
            return []
        return values[:index + 1]


def infer_tab_name_from_filename(filename):
    return os.path.basename(filename).split('.')[0]


class SingleTableMixin(AbstractTableSetManager):

    def __init__(self, filename: str, tab_name: Optional[str] = None, **kwargs):
        self._tab_name = tab_name or infer_tab_name_from_filename(filename)
        super().__init__(filename=filename, **kwargs)

    def tab_names(self, order: Optional[List[str]] = None) -> List[str]:
        return [self._tab_name]


class InsertsManager(BasicTableSetManager):

    ALLOWED_FILE_EXTENSIONS = []

    def _parse_inserts_data(self, filename: str) -> AnyJsonData:
        raise NotImplementedError(f"._parse_inserts_dataa(...) is not implemented for {self.__class__.__name__}.")  # noQA

    def _load_inserts_data(self, filename: str) -> TabbedSheetData:
        raw_data: AnyJsonData = self._parse_inserts_data(filename)
        tabbed_inserts: AnyJsonData = self._wrap_inserts_data(filename, raw_data)
        if (not isinstance(tabbed_inserts, dict)
                or not all(isinstance(tab_name, str) for tab_name in tabbed_inserts.keys())
                or not all(isinstance(data, list) and all(isinstance(datum, dict) for datum in data)
                           for data in tabbed_inserts.values())):
            raise ValueError(f"Data in {filename} is not of type TabbedSheetData (Dict[str, List[dict]]).")
        tabbed_inserts: TabbedSheetData  # we've just checked that
        return tabbed_inserts

    @classmethod
    def _wrap_inserts_data(cls, filename: str, data: AnyJsonData) -> AnyJsonData:
        ignored(filename)
        return data

    def tab_names(self, order: Optional[List[str]] = None) -> List[str]:
        return list(self.content_by_tab_name.keys())

    def _get_reader_agent(self) -> Any:
        return self

    @classmethod
    def extract_tabbed_headers(cls, data: TabbedSheetData) -> TabbedHeaders:
        result: TabbedHeaders = {}
        for tab, rows in data.items():
            if rows:
                # Data is homogeneous, so whatever the headers for the first row should be the same for all
                headers: List[str] = list(rows[0].keys())
            else:
                # If there's no data in the tab, there are also no headers
                headers: List[str] = []
            result[tab] = headers
        return result

    def load_content(self, sheet_order: Optional[List[str]] = None) -> Dict[str, AnyJsonData]:
        data = self._load_inserts_data(self.filename)
        self.content_by_tab_name = data
        self.headers_by_tab_name = self.extract_tabbed_headers(data)
        return self.content_by_tab_name


class SimpleInsertsMixin(SingleTableMixin):

    def _wrap_inserts_data(self, filename: str, data: AnyJsonData) -> TabbedSheetData:
        if (not isinstance(data, list)
                or not all(isinstance(datum, dict) for datum in data)):
            raise ValueError(f"Data in {filename} is not of type SheetData (List[dict]).")
        return {self._tab_name: data}


class JsonInsertsMixin:

    @classmethod
    def _parse_inserts_data(cls, filename: str) -> AnyJsonData:
        return json.load(open_unicode_text_input_file_respecting_byte_order_mark(filename))


@TABLE_SET_MANAGER_REGISTRY.register()
class TabbedJsonInsertsManager(JsonInsertsMixin, InsertsManager):

    ALLOWED_FILE_EXTENSIONS = [".tabs.json"]  # If you want them all in one family, use this extension


@TABLE_SET_MANAGER_REGISTRY.register()
class SimpleJsonInsertsManager(SimpleInsertsMixin, JsonInsertsMixin, InsertsManager):

    ALLOWED_FILE_EXTENSIONS = [".json"]


class YamlInsertsMixin:

    def _parse_inserts_data(self, filename) -> AnyJsonData:
        return yaml.safe_load(open_unicode_text_input_file_respecting_byte_order_mark(filename))


@TABLE_SET_MANAGER_REGISTRY.register()
class TabbedYamlInsertsManager(YamlInsertsMixin, InsertsManager):

    ALLOWED_FILE_EXTENSIONS = [".tabs.yaml"]

    def _parse_inserts_data(self, filename) -> AnyJsonData:
        return yaml.safe_load(open_unicode_text_input_file_respecting_byte_order_mark(filename))


@TABLE_SET_MANAGER_REGISTRY.register()
class SimpleYamlInsertsManager(SimpleInsertsMixin, YamlInsertsMixin, InsertsManager):

    ALLOWED_FILE_EXTENSIONS = [".yaml"]


@TABLE_SET_MANAGER_REGISTRY.register()
class SimpleJsonLinesInsertsManager(SimpleInsertsMixin, InsertsManager):

    ALLOWED_FILE_EXTENSIONS = [".jsonl"]

    def _parse_inserts_data(self, filename: str) -> AnyJsonData:
        return [line for line in JsonLinesReader(open_unicode_text_input_file_respecting_byte_order_mark(filename))]


@TABLE_SET_MANAGER_REGISTRY.register(regexp="^(.*/)?(|[^/]*[-_])inserts/?$")
class InsertsDirectoryManager(InsertsManager):

    ALLOWED_FILE_EXTENSIONS = []

    def _parse_inserts_data(self, filename: str) -> AnyJsonData:
        if not os.path.isdir(filename):
            raise LoadArgumentsError(f"{filename} is not the name of an inserts directory.")
        tab_files = glob.glob(os.path.join(filename, "*.json"))
        data = {}
        for tab_file in tab_files:
            tab_content = json.load(open_unicode_text_input_file_respecting_byte_order_mark(tab_file))
            # Here we don't use os.path.splitext because we want to split on the first dot.
            # e.g., for foo.bar.baz, return just foo
            #       this allows names like ExperimentSet.tab.json that might need to use multi-dot suffixes
            #       for things unrelated to the tab name.
            tab_name = os.path.basename(tab_file).split('.')[0]
            data[tab_name] = tab_content
        return data


@TABLE_SET_MANAGER_REGISTRY.register()
class CsvManager(SingleTableMixin, FlattenedTableSetManager):
    """
    This implements the mechanism to get a series of rows out of the sheet in a csv file,
    returning a result that still looks like there could have been multiple tabs.
    """

    ALLOWED_FILE_EXTENSIONS = ['.csv']

    def __init__(self, filename: str, escaping: Optional[bool] = None, **kwargs):
        super().__init__(filename=filename, **kwargs)
        self.escaping: bool = escaping or False

    def _get_reader_agent(self) -> CsvReader:
        return self._get_reader_agent_for_filename(self.filename)

    @classmethod
    def _get_reader_agent_for_filename(cls, filename) -> CsvReader:
        return csv.reader(open_unicode_text_input_file_respecting_byte_order_mark(filename))

    PAD_TRAILING_TABS = True

    def _raw_row_generator_for_tab_name(self, tab_name: str) -> Iterable[SheetRow]:
        headers = self.tab_headers(tab_name)
        n_headers = len(headers)
        for row_data in self.reader_agent:
            if self.PAD_TRAILING_TABS:
                row_data = pad_to(n_headers, row_data, padding='')
            yield row_data

    def _create_tab_processor_state(self, tab_name: str) -> Headers:
        headers: Optional[Headers] = self.headers_by_tab_name.get(tab_name)
        if headers is None:
            self.headers_by_tab_name[tab_name] = headers = self.reader_agent.__next__()
        return headers

    @classmethod
    def _escape_cell_text(cls, cell_text):
        if '\\' in cell_text:
            return expand_string_escape_sequences(cell_text)
        else:
            return cell_text

    def _process_row(self, tab_name: str, headers: Headers, row_data: SheetRow) -> AnyJsonData:
        ignored(tab_name)
        if self.escaping:
            return {headers[i]: self.parse_cell_value(self._escape_cell_text(cell_text), override_prefer_number=False)
                    for i, cell_text in enumerate(row_data)}
        else:
            return {headers[i]: self.parse_cell_value(cell_text, override_prefer_number=False)
                    for i, cell_text in enumerate(row_data)}


@TABLE_SET_MANAGER_REGISTRY.register()
class TsvManager(CsvManager):
    """
    TSV files are just CSV files with tabs instead of commas as separators.
    (We do not presently handle any escaping of strange characters. May need to add handling for backslash escaping.)
    """
    ALLOWED_FILE_EXTENSIONS = ['.tsv', '.tsv.txt']

    @classmethod
    def _get_reader_agent_for_filename(cls, filename) -> CsvReader:
        return csv.reader(open_unicode_text_input_file_respecting_byte_order_mark(filename), delimiter='\t')


def _do_shell_command(command, cwd=None):
    # This might need to be more elaborate, but hopefully it will do for now. -kmp 11-Sep-2023
    subprocess.check_output(command, cwd=cwd)


@contextlib.contextmanager
def maybe_unpack(filename):  # Maybe move to another module
    """
    If necessary, unpack a file that is zipped and/or tarred, yielding the name of the file (unpacked or not).
    """
    unpackables = ['.tar.gz', '.tar', '.tgz', '.gz', '.zip']
    ext = None
    for unpackable in unpackables:
        if filename.endswith(unpackable):
            ext = unpackable
            break
    if not ext:
        yield filename
        return
    if not os.path.exists(filename):
        # We don't bother to raise this error if we're not planning to do any unpacking.
        # The caller can decide if/when such errors are needed in that case.
        # But if we are going to have to move bits around, they'll need to actually be there.
        # -kmp 12-Sep-2023
        raise ValueError(f"The file {filename!r} does not exist.")
    target_base_part = remove_suffix(ext, os.path.basename(filename), required=True)
    target_ext = '.tar.gz' if ext == '.tgz' else ext
    with TemporaryDirectory() as temp_dir:
        temp_base = os.path.join(temp_dir, target_base_part)
        temp_filename = temp_base + target_ext
        _do_shell_command(['cp', filename, temp_filename])
        if temp_filename.endswith('.gz'):
            _do_shell_command(['gunzip', temp_filename], cwd=temp_dir)
            temp_filename = remove_suffix('.gz', temp_filename)
        elif temp_filename.endswith(".zip"):
            _do_shell_command(['unzip', temp_filename], cwd=temp_dir)
            temp_filename = remove_suffix('.zip', temp_filename)
        if temp_filename.endswith(".tar"):
            _do_shell_command(['tar', '-xf', temp_filename], cwd=temp_dir)
            tar_file = temp_filename
            temp_filename = remove_suffix(".tar", temp_filename, required=True)
            if not os.path.isdir(temp_filename):
                raise Exception(f"{tar_file} didn't unpack to a dir: {temp_filename}")
        # print(f"Unpacked {filename} to {temp_filename}")
        yield temp_filename


class TableSetManager(AbstractTableSetManager):
    """
    This class will open a .xlsx or .csv file and load its content in our standard format.
    (See more detailed description in AbstractTableManager.)
    """

    COMPRESSION_EXTENSIONS = ['.gz', '.tgz', '.tar.gz']

    @classmethod
    def create_implementation_manager(cls, filename: str, **kwargs) -> AbstractTableSetManager:

        reader_agent_class = TABLE_SET_MANAGER_REGISTRY.manager_for_filename(filename)
        # This is a bad forward reference in current refactor, but also may be testing for something we don't need
        # to worry about anymore. -kmp 11-Sep-2023
        #
        # if issubclass(reader_agent_class, AbstractItemManager):
        #     raise ValueError(f"TableSetManager unexpectedly found reader agent class {reader_agent_class}.")
        reader_agent = reader_agent_class(filename=filename, **kwargs)
        return reader_agent

    @classmethod
    def load(cls, filename: str, tab_name: Optional[str] = None, escaping: Optional[bool] = None,
             **kwargs) -> TabbedSheetData:
        """
        Given a filename and various options
        """
        annotated_content = cls.load_annotated(filename=filename, tab_name=tab_name, escaping=escaping, **kwargs)
        content: TabbedSheetData = annotated_content['content']
        return content

    @classmethod
    def load_annotated(cls, filename: str, tab_name: Optional[str] = None, escaping: Optional[bool] = None,
                       retain_empty_properties: bool = False, sheet_order: Optional[List[str]] = None,
                       **kwargs) -> Dict:
        """
        Given a filename and various options
        """
        orig_filename = filename
        with maybe_unpack(filename) as filename:
            manager = cls.create_implementation_manager(filename=filename, tab_name=tab_name, escaping=escaping,
                                                        **kwargs)
            content: TabbedSheetData = manager.load_content(sheet_order)
            return {
                'filename': filename,
                'content': content,
                'tab_name': tab_name,
                'escaping': escaping,
                'singleton': isinstance(manager, SingleTableMixin),
                'flattened': isinstance(manager, FlattenedTableSetManager),
                'packed': orig_filename != filename,  # tar or zip file that had to be unpacked somehow
            }


load_table_set = TableSetManager.load
load_table_annotated = TableSetManager.load_annotated



from copy import deepcopy
from functools import lru_cache
import re
from typing import Any, Callable, List, Optional, Tuple, Type, Union
from dcicutils.data_readers import RowReader
from dcicutils.misc_utils import create_readonly_object
from dcicutils.portal_utils import Portal
from dcicutils.schema_utils import Schema

PortalObject = Type["PortalObject"]  # Forward type reference for type hints.


class PortalObject:

    _PROPERTY_DELETION_SENTINEL = RowReader.CELL_DELETION_SENTINEL

    def __init__(self, data: dict, portal: Portal = None,
                 schema: Optional[Union[dict, Schema]] = None, type: Optional[str] = None) -> None:
        self._data = data if isinstance(data, dict) else {}
        self._portal = portal if isinstance(portal, Portal) else None
        self._schema = schema if isinstance(schema, dict) else (schema.data if isinstance(schema, Schema) else None)
        self._type = type if isinstance(type, str) else ""

    @property
    def data(self) -> dict:
        return self._data

    @property
    def portal(self) -> Optional[Portal]:
        return self._portal

    @property
    @lru_cache(maxsize=1)
    def type(self) -> str:
        return self._type or Portal.get_schema_type(self._data) or (Schema(self._schema).type if self._schema else "")

    @property
    @lru_cache(maxsize=1)
    def types(self) -> Optional[List[str]]:
        return [self._type] if self._type else Portal.get_schema_types(self._data)

    @property
    @lru_cache(maxsize=1)
    def uuid(self) -> Optional[str]:
        return self._data.get("uuid") if isinstance(self._data, dict) else None

    @property
    @lru_cache(maxsize=1)
    def schema(self) -> Optional[dict]:
        return self._schema if self._schema else (self._portal.get_schema(self.type) if self._portal else None)

    def copy(self) -> PortalObject:
        return PortalObject(deepcopy(self.data), portal=self.portal, type=self.type)

    @property
    @lru_cache(maxsize=1)
    def identifying_properties(self) -> Optional[List[str]]:
        """
        Returns the list of all identifying property names of this Portal object which actually have values.
        Implicitly include "uuid" and "identifier" properties as identifying properties if they are actually
        properties in the object schema, and favor these (first); defavor "aliases"; no other ordering defined.
        """
        if not (schema := self.schema) or not (schema_identifying_properties := schema.get("identifyingProperties")):
            return None
        identifying_properties = []
        for identifying_property in schema_identifying_properties:
            if identifying_property not in ["uuid", "identifier", "aliases"]:
                if self._data.get(identifying_property):
                    identifying_properties.append(identifying_property)
        if self._data.get("identifier"):
            identifying_properties.insert(0, "identifier")
        if self._data.get("uuid"):
            identifying_properties.insert(0, "uuid")
        if "aliases" in schema_identifying_properties and self._data.get("aliases"):
            identifying_properties.append("aliases")
        return identifying_properties or None

    @lru_cache(maxsize=8192)
    def lookup(self, raw: bool = False,
               ref_lookup_strategy: Optional[Callable] = None) -> Tuple[Optional[PortalObject], Optional[str], int]:
        nlookups = 0
        first_identifying_path = None
        try:
            if identifying_paths := self._get_identifying_paths(ref_lookup_strategy=ref_lookup_strategy):
                for identifying_path in identifying_paths:
                    if not first_identifying_path:
                        first_identifying_path = identifying_path
                    nlookups += 1
                    if (value := self._portal.get(identifying_path, raw=raw)) and (value.status_code == 200):
                        return (
                            PortalObject(value.json(), portal=self._portal, type=self.type if raw else None),
                            identifying_path,
                            nlookups
                        )
        except Exception:
            pass
        return None, first_identifying_path, nlookups

    def compare(self, value: Union[dict, PortalObject],
                consider_refs: bool = False, resolved_refs: List[dict] = None) -> Tuple[dict, int]:
        if consider_refs and isinstance(resolved_refs, list):
            normlized_portal_object, nlookups = self._normalized_refs(refs=resolved_refs)
            this_data = normlized_portal_object.data
        else:
            this_data = self.data
            nlookups = 0
        if isinstance(value, PortalObject):
            comparing_data = value.data
        elif isinstance(value, dict):
            comparing_data = value
        else:
            return {}, nlookups
        return PortalObject._compare(this_data, comparing_data), nlookups

    @staticmethod
    def _compare(a: Any, b: Any, _path: Optional[str] = None) -> dict:
        def diff_creating(value: Any) -> object:  # noqa
            return create_readonly_object(value=value,
                                          creating_value=True, updating_value=None, deleting_value=False)
        def diff_updating(value: Any, updating_value: Any) -> object:  # noqa
            return create_readonly_object(value=value,
                                          creating_value=False, updating_value=updating_value, deleting_value=False)
        def diff_deleting(value: Any) -> object:  # noqa
            return create_readonly_object(value=value,
                                          creating_value=False, updating_value=None, deleting_value=True)
        diffs = {}
        if isinstance(a, dict) and isinstance(b, dict):
            for key in a:
                path = f"{_path}.{key}" if _path else key
                if key not in b:
                    if a[key] != PortalObject._PROPERTY_DELETION_SENTINEL:
                        diffs[path] = diff_creating(a[key])
                else:
                    diffs.update(PortalObject._compare(a[key], b[key], _path=path))
        elif isinstance(a, list) and isinstance(b, list):
            # Ignore order of array elements; not absolutely technically correct but suits our purpose.
            for index in range(len(a)):
                path = f"{_path or ''}#{index}"
                if not isinstance(a[index], dict) and not isinstance(a[index], list):
                    if a[index] not in b:
                        if a[index] != PortalObject._PROPERTY_DELETION_SENTINEL:
                            if index < len(b):
                                diffs[path] = diff_updating(a[index], b[index])
                            else:
                                diffs[path] = diff_creating(a[index])
                        else:
                            if index < len(b):
                                diffs[path] = diff_deleting(b[index])
                elif index < len(b):
                    diffs.update(PortalObject._compare(a[index], b[index], _path=path))
                else:
                    diffs[path] = diff_creating(a[index])
        elif a != b:
            if a == PortalObject._PROPERTY_DELETION_SENTINEL:
                diffs[_path] = diff_deleting(b)
            else:
                diffs[_path] = diff_updating(a, b)
        return diffs

    @lru_cache(maxsize=1)
    def _get_identifying_paths(self, ref_lookup_strategy: Optional[Callable] = None) -> Optional[List[str]]:
        """
        Returns a list of the possible Portal URL paths identifying this Portal object.
        """
        identifying_paths = []
        if not (identifying_properties := self.identifying_properties):
            if self.uuid:
                if self.type:
                    identifying_paths.append(f"/{self.type}/{self.uuid}")
                identifying_paths.append(f"/{self.uuid}")
            return identifying_paths
        for identifying_property in identifying_properties:
            if identifying_value := self._data.get(identifying_property):
                if identifying_property == "uuid":
                    if self.type:
                        identifying_paths.append(f"/{self.type}/{identifying_value}")
                    identifying_paths.append(f"/{identifying_value}")
                # For now at least we include the path both with and without the schema type component,
                # as for some identifying values, it works (only) with, and some, it works (only) without.
                # For example: If we have FileSet with "accession", an identifying property, with value
                # SMAFSFXF1RO4 then /SMAFSFXF1RO4 works but /FileSet/SMAFSFXF1RO4 does not; and
                # conversely using "submitted_id", also an identifying property, with value
                # UW_FILE-SET_COLO-829BL_HI-C_1 then /UW_FILE-SET_COLO-829BL_HI-C_1 does
                # not work but /FileSet/UW_FILE-SET_COLO-829BL_HI-C_1 does work.
                elif isinstance(identifying_value, list):
                    for identifying_value_item in identifying_value:
                        if self.type:
                            identifying_paths.append(f"/{self.type}/{identifying_value_item}")
                        identifying_paths.append(f"/{identifying_value_item}")
                else:
                    # TODO: Import from somewhere ...
                    lookup_options = 0
                    if schema := self.schema:
                        # TODO: Hook into the ref_lookup_strategy thing in structured_data to make
                        # sure we check accession format (since it does not have a pattern).
                        if callable(ref_lookup_strategy):
                            lookup_options, ref_validator = ref_lookup_strategy(
                                self.type, schema, identifying_value)
                            if callable(ref_validator):
                                if ref_validator(schema, identifying_property, identifying_value) is False:
                                    continue
                        if pattern := schema.get("properties", {}).get(identifying_property, {}).get("pattern"):
                            if not re.match(pattern, identifying_value):
                                # If this identifying value is for a (identifying) property which has a
                                # pattern, and the value does NOT match the pattern, then do NOT include
                                # this value as an identifying path, since it cannot possibly be found.
                                continue
                    if not lookup_options:
                        lookup_options = Portal.LOOKUP_DEFAULT
                    if Portal.is_lookup_root_first(lookup_options):
                        identifying_paths.append(f"/{identifying_value}")
                    if Portal.is_lookup_specified_type(lookup_options) and self.type:
                        identifying_paths.append(f"/{self.type}/{identifying_value}")
                    if Portal.is_lookup_root(lookup_options) and not Portal.is_lookup_root_first(lookup_options):
                        identifying_paths.append(f"/{identifying_value}")
                    if Portal.is_lookup_subtypes(lookup_options):
                        for subtype_name in self._portal.get_schema_subtype_names(self.type):
                            identifying_paths.append(f"/{subtype_name}/{identifying_value}")
        return identifying_paths or None

    def _normalized_refs(self, refs: List[dict]) -> Tuple[PortalObject, int]:
        """
        Same as _normalize_ref but does NOT make this change to this Portal object IN PLACE,
        rather it returns a new instance of this Portal object wrapped in a new PortalObject.
        """
        portal_object = self.copy()
        nlookups = portal_object._normalize_refs(refs)
        return portal_object, nlookups

    def _normalize_refs(self, refs: List[dict]) -> int:
        """
        Turns any (linkTo) references which are paths (e.g. /SubmissionCenter/uwsc_gcc) within this
        object IN PLACE into the uuid style reference (e.g. d1b67068-300f-483f-bfe8-63d23c93801f),
        based on the given "refs" list which is assumed to be a list of dictionaries, where each
        contains a "path" and a "uuid" property; this list is typically (for our first usage of
        this function) the value of structured_data.StructuredDataSet.resolved_refs_with_uuid.
        Changes are made to this Portal object IN PLACE; use _normalized_refs function to make a copy.
        If there are no "refs" (None or empty) or if the speicified reference is not found in this
        list then the references will be looked up via Portal calls (via Portal.get_metadata).
        """
        _, nlookups = PortalObject._normalize_data_refs(self.data, refs=refs, schema=self.schema, portal=self.portal)
        return nlookups

    @staticmethod
    def _normalize_data_refs(value: Any, refs: List[dict], schema: dict,
                             portal: Portal, _path: Optional[str] = None) -> Tuple[Any, int]:
        nlookups = 0
        if not value or not isinstance(schema, dict):
            return value, nlookups
        if isinstance(value, dict):
            for key in value:
                path = f"{_path}.{key}" if _path else key
                value[key], nlookups = PortalObject._normalize_data_refs(value[key], refs=refs,
                                                                         schema=schema, portal=portal, _path=path)
        elif isinstance(value, list):
            for index in range(len(value)):
                path = f"{_path or ''}#{index}"
                value[index], nlookups = PortalObject._normalize_data_refs(value[index], refs=refs,
                                                                           schema=schema, portal=portal, _path=path)
        elif value_type := Schema.get_property_by_path(schema, _path):
            if link_to := value_type.get("linkTo"):
                ref_path = f"/{link_to}/{value}"
                if not isinstance(refs, list):
                    refs = []
                if ref_uuids := [ref.get("uuid") for ref in refs if ref.get("path") == ref_path]:
                    ref_uuid = ref_uuids[0]
                else:
                    ref_uuid = None
                if ref_uuid:
                    return ref_uuid, nlookups
                # Here our (linkTo) reference appears not to be in the given refs; if these refs came
                # from structured_data.StructuredDataSet.resolved_refs_with_uuid (in the context of
                # smaht-submitr, which is the typical/first use case for this function) then this could
                # be because the reference was to an internal object, i.e. another object existing within
                # the data/spreadsheet being submitted. In any case, we don't have the associated uuid
                # so let us look it up here.
                if isinstance(portal, Portal):
                    nlookups += 1
                    if ((ref_object := portal.get_metadata(ref_path, raise_exception=False)) and
                        (ref_uuid := ref_object.get("uuid"))):  # noqa
                        return ref_uuid, nlookups
        return value, nlookups



import datetime
import re

from .misc_utils import ignored, capitalize1
from typing import Union, Optional


class EnglishUtils:
    """
    In most cases you can get away without using this class, but it's here in case customization is needed.
    The usual interfaces are available as separate functions.
    """

    SECOND = 1
    MINUTE = 60 * SECOND
    HOUR = 60 * MINUTE
    DAY = 24 * HOUR
    WEEK = 7 * DAY

    _TIME_UNITS = ((WEEK, "week"), (DAY, "day"), (HOUR, "hour"), (MINUTE, "minute"))

    _SPECIAL_PLURALS = {
        "radius": "radii",
        "spectrum": "spectra",
        "ovum": "ova",
        "deer": "deer",
        "fish": "fish",
        "goose": "geese",
        "sheep": "sheep",
        "tooth": "teeth",
        "foot": "feet",
        "ox": "oxen",
        'datum': 'data',
        'metadatum': 'metadata',
        'sis': 'sisses',
        'stepsis': 'stepsisses',
    }

    @classmethod
    def _special_case_plural(cls, word: str) -> str:
        """Returns either a special case plural of its argument, or the empty string if it doesn't know."""
        return cls._SPECIAL_PLURALS.get(word, "")

    # There are some other rules in https://languagetool.org/insights/post/plural-nouns/
    # that we might want to consider at some point.
    _ENDS_IN_FE = re.compile(r".*[aeiou]fe$", flags=re.IGNORECASE)
    _ENDS_IN_F = re.compile(r".*[aeoul]f$", flags=re.IGNORECASE)
    _ENDS_IN_MAN = re.compile(r".*man$", flags=re.IGNORECASE)
    _ENDS_IN_HUMAN = re.compile(r".*human$", flags=re.IGNORECASE)
    _ENDS_IN_CHILD = re.compile(r".*child$", flags=re.IGNORECASE)
    _ENDS_IN_SIS = re.compile(r".*sis$", flags=re.IGNORECASE)
    _ENDS_IN_VOWEL_Z = re.compile(r".*[aeiou]z$", flags=re.IGNORECASE)
    _ENDS_IN_XSZ_OR_SH_OR_CH = re.compile(r".*([xsz]|[cs]h)$", flags=re.IGNORECASE)
    _ENDS_IN_NONVOWEL_Y = re.compile(r".*[^aeiou]y$", flags=re.IGNORECASE)

    @classmethod
    def _adjust_ending(cls, word, strip_chars, add_suffix):
        return (word[:-strip_chars] if strip_chars else word) + add_suffix

    _COMPOUND_PLURAL_SIMPLE_PREPOSITIONS = [
        'about', 'at', 'between', 'by', 'for', 'from', 'in', 'of', 'on', 'to', 'with'
    ]

    # Phrases like 'using', 'used by', 'used in', etc. function similarly to prepositions when doing pluralization
    # in that the plural of 'a variant referencing a gene' would be 'variants referencing genes', just as
    # the plural of 'a gene referenced by a variant' would be 'genes referenced by variants'.

    _COMPOUND_PLURAL_PSEUDO_CONNECTIVES = [
        'using', 'containing', 'including', 'referencing', 'naming',
        'used by', 'contained by', 'included by', 'referenced by', 'named by',
        'used in', 'contained in', 'included in', 'referenced in', 'named in',
    ]

    _COMPOUND_PLURAL_PREPOSITIONS = _COMPOUND_PLURAL_SIMPLE_PREPOSITIONS + _COMPOUND_PLURAL_PSEUDO_CONNECTIVES

    _NOUN_WITH_PREPOSITIONAL_ATTACHMENT = re.compile(
        # Note use of *? to get minimal match in group1, so that we'll find the first preposition, not a later one
        # Specifically, the use of "*?" makes us treat multiple prepositions like a-of-b-of-c by recursing as
        # (a)(-)(of)(-)(b-of-c). If we used "*" instead of "*?", we would get (a-of-b)(-)(of)(-)(c).
        # If we were only doing hyphenated items, it wouldn't matter a whole lot.
        # But in the words part, it matters a great deal because with spaces and hyphens interleaved, there is a
        # risk of parsing "a-of-b of c-of-d" as (a-of-b of c)(-)(of)(-)(d), whereas I think (a-of-b)( )(of)( )(c-of-d)
        # is more likely to get a better parse. -kmp 29-Aug-2021
        f"""^(?:
                 # Either these 5 will match (and the next 5 will not)
                 # to ONLY match a hyphenated compound word like son-in-law as (son)(-)(in)(-)(law)
                 ([a-z][a-z-]*?)  # shortest possible (i.e., first) block of hyphenated words preceding "-<prep>-"
                 ([-])            # pre-preposition-hyphen to make sure it's part of the same compound.
                 ({'|'.join(_COMPOUND_PLURAL_PREPOSITIONS).replace(' ', '[ ]')}) # matches the prep, as (about|at|...)
                 ([-])            # hyphen on the other side of prep
                 ([a-z-]*)        # we're less fussy about this. It could contain more preps, for example.
              |
                 # or these 5 will match (and the previous 5 will not)...
                 # to match shortest (i.e., first) words (included hyphenated words) leading to " <prep> "
                 # so 'the son-in-law of the proband' becomes (the son-in-law)( )(of)( )(the proband)
                 ([a-z]        # first token starts with an alphabetic,
                  [a-z- ]*?    # matches any number of words, which may be hyphenated,
                  [a-z])       # and ends in an alphabetic
                 ([ ]+)        # unlike with hyphenation, any number of pre-<prep> spaces is ok
                 ({'|'.join(_COMPOUND_PLURAL_PREPOSITIONS).replace(' ', '[ ]')}) # matches the prep, as (about|at|...)
                 ([ ]+)        # any amount of whitespace on the other side
                 (.*)          # anything else that follows first space-delimeted preposition
             )$""",

        re.IGNORECASE | re.VERBOSE)
    _INDEFINITE_NOUN_REF = re.compile(
        f"^an?[ -]+([^ -].*)$",
        re.IGNORECASE)

    _NOUN_WITH_CLAUSE_QUALIFIER = re.compile("^(.*[^,])(,|)[ ]+(that|which|while)[ ]+(.*)$", re.IGNORECASE)
    _IS_QUALIFIER = re.compile("^(is|was|has)[ ]+(.*)$", re.IGNORECASE)

    @classmethod
    def string_pluralize(cls, word: str, allow_some=False) -> str:
        """
        Returns the probable plural of the given word.
        This is an ad hoc string pluralizer intended for situations where being mostly right is good enough.
        e.g., string_pluralize('sample') => 'sample'
              string_pluralize('community') => 'communities'
        """

        qualifier_suffix = ""

        charn = word[-1]
        capitalize = word[0].isupper()
        upcase = word.isupper()  # capitalize and not any(ch.islower() for ch in word)

        qual_matched = cls._NOUN_WITH_CLAUSE_QUALIFIER.match(word)
        if qual_matched:
            qualified, comma, connective, qualifier = qual_matched.groups()
            word = qualified
            is_matched = cls._IS_QUALIFIER.match(qualifier)
            if is_matched:
                verb, qualifying_adj = is_matched.groups()
                orig_verb = verb
                verb = {'is': 'are', 'was': 'were', 'has': 'have'}.get(verb.lower(), verb)
                if orig_verb[0].isupper():
                    if orig_verb[-1].isupper():
                        verb = verb.upper()
                    else:
                        verb = verb.capitalize()
                qualifier = f"{verb} {qualifying_adj}"
            # Continue to other things after making a verb adjustment
            qualifier_suffix = f"{comma} {connective} {qualifier}"

        # Convert 'a foo' to just 'foo' prior to pluralization. It's pointless to return 'a apples' or 'an apples'.
        # Arguably, we _could_ return 'some apples'
        indef_matched = cls._INDEFINITE_NOUN_REF.match(word)
        if indef_matched:
            word = indef_matched.group(1)
            if allow_some:
                prefix = "SOME " if upcase else ("Some " if capitalize else "some ")
                return prefix + cls.string_pluralize(word) + qualifier_suffix

        prep_matched = cls._NOUN_WITH_PREPOSITIONAL_ATTACHMENT.match(word)
        if prep_matched:
            groups = prep_matched.groups()
            word, prep_spacing1, prep, prep_spacing2, prep_obj = groups[0:5] if groups[0] is not None else groups[5:10]
            indef_matched = cls._INDEFINITE_NOUN_REF.match(prep_obj)
            if indef_matched:
                # It's important to do this before calling ourselves recursively to avoid getting 'some' in prep phrases
                prep_obj = indef_matched.group(1)
                prep_obj = cls.string_pluralize(prep_obj)
            return cls.string_pluralize(word) + prep_spacing1 + prep + prep_spacing2 + prep_obj + qualifier_suffix

        result = cls._special_case_plural(word)
        if result:
            return result + qualifier_suffix

        if cls._ENDS_IN_FE.match(word):
            result = cls._adjust_ending(word, 2, "ves")
        elif cls._ENDS_IN_F.match(word):
            result = cls._adjust_ending(word, 1, "ves")
        elif cls._ENDS_IN_MAN.match(word) and not cls._ENDS_IN_HUMAN.match(word):
            result = cls._adjust_ending(word, 2, "e" + charn)
        elif cls._ENDS_IN_CHILD.match(word):
            result = cls._adjust_ending(word, 0, "ren")
        elif cls._ENDS_IN_SIS.match(word):
            result = cls._adjust_ending(word, 2, "es")
        elif cls._ENDS_IN_VOWEL_Z.match(word):
            result = cls._adjust_ending(word, 0, "zes")
        elif cls._ENDS_IN_XSZ_OR_SH_OR_CH.match(word):
            result = cls._adjust_ending(word, 0, "es")
        elif cls._ENDS_IN_NONVOWEL_Y.match(word):
            result = cls._adjust_ending(word, 1, "ies")
        else:
            result = cls._adjust_ending(word, 0, "s")

        if upcase:
            return result.upper() + qualifier_suffix
        elif capitalize:
            return result.capitalize() + qualifier_suffix
        else:
            return result + qualifier_suffix

    _USE_AN = {}

    _PREFIX_PATTERN_FOR_A = re.compile("^(%s)" % "|".join({
        "[^aeioux]",  # Consonants other than x need 'a' (bicycle, dog, etc.)
        "x[aeiouy]",  # x followed by any vowel makes it pronounceable like a consonant (xylophone), so needs 'a'
        "uni([^aeiuym]|[aeiuy][^aeiy])",  # things starting with "uni" are pronounced like "yuni", so need "a"
    }), flags=re.IGNORECASE)

    @classmethod
    def select_a_or_an(cls, word):
        """
        Uses a heuristic to try to select the appropriate article ("a" or "an") for a given English noun.
        select_a_or_an("gene") => 'a'
        select_a_or_an("accession") => 'an'
        """

        return "a" if cls._PREFIX_PATTERN_FOR_A.match(word) else "an"

    @classmethod
    def a_or_an(cls, word):
        """
        Heuristically attaches either "a" or "an" to a given English noun.
        a_or_an("gene") => "a gene"
        a_or_an("accession") => "an accession"
        """
        article = cls.select_a_or_an(word)
        return "%s %s" % (article, word)

    @classmethod
    def maybe_pluralize(cls, n, thing):
        """
        Given a number and a noun, returns the singular or plural of the noun as appropriate.

        The number may simply be a collection (list, tuple, etc.), in which case len is used.

        NOTE: The number itself does not appear in the return value.
              It is used only to determine if pluralization is needed.
        """

        if isinstance(n, (list, tuple, set, dict)):
            n = len(n)
        return thing if n == 1 else cls.string_pluralize(thing)

    @classmethod
    def n_of(cls, n, thing, num_format=None):
        """
        Given a number and a noun, returns the name for that many of that noun.

        Examples:

            >>> n_of(7, "variant")
            '7 variants'
            >>> n_of(1, "accession")
            '1 accession'
            >>> n_of(['alpha', 'beta', 'gamma'], 'Greek letter')
            '3 Greek letters'
        """
        if isinstance(n, (list, tuple, set, dict)):
            n = len(n)
        display_n = n
        if num_format:
            res = num_format(n, thing)
            if res:
                display_n = res
        return "%s %s" % (display_n, thing if n == 1 else cls.string_pluralize(thing))

    @classmethod
    def must_be_one_of(cls, items, *, possible: Union[bool, str] = True, kind: str = "option", quote=False,
                       capitalize=True, joiner=None, **joiner_options):
        """
        Constructs a sentence that complains about a given quantity not being among a given set of options.

        This is intended to be useful in error messages to enumerate a set of values, usually but not necessarily
        strings, that had been expected but not received. For example:

        >>> must_be_one_of([])
        "There are no possible options."
        >>> must_be_one_of(['foo'])
        "The only possible option is foo."
        >>> must_be_one_of(['foo', 'bar'])
        "Possible options are foo and bar."
        >>> must_be_one_of(['foo', 'bar', 'baz'])
        "Possible options are foo, bar and baz."

        :param items: the items to enumerate
        :param possible: whether to use the word 'possible' before the given kind (default True), or a string to use
        :param kind: the kind of items being enumerated (default "option")
        :param quote: whether to put quotes around each option
        :param capitalize: whether to capitalize the first letter of the sentence (default True)
        :param joiner: the joining function to join the items (default if None is just a commas-separated list)
        :param joiner_options: additional (keyword) options to be used with a joiner function if one is supplied
        """

        n = len(items)
        maybe_adj = ""
        if possible:
            if possible is True:
                possible = "possible"
            maybe_adj = possible + " "
        if not joiner:
            joiner = cls.conjoined_list
        if quote:
            # First force to a string, so we don't call the item's repr, then use repr to add quotation marks.
            items = [repr(str(item)) for item in items]
        if n == 0:
            kinds = cls.string_pluralize(kind)
            result = f"there are no {maybe_adj}{kinds}."
        elif n == 1:
            [item] = items
            result = f"the only {maybe_adj}{kind} is {item}."
        else:
            kinds = cls.string_pluralize(kind)
            options = joiner(items, **joiner_options)
            result = f"{maybe_adj}{kinds} are {options}."
        if capitalize:
            result = capitalize1(result)
        return result

    _BE_TENSES = {
        'past': {'singular': 'was', 'plural': 'were'},
        'present': {'singular': 'is', 'plural': 'are'},
        'past-perfect': {'singular': 'has been', 'plural': 'have been'},
    }

    _MODALS = {'can', 'could', 'may', 'might', 'must', 'shall', 'should', 'will', 'would'}

    @classmethod
    def _conjugate_be(cls, count, tense):
        """
        Returns the conjugation of the verb 'to be' for a subject that has a given count and tense.

        For example:

        >>> EnglishUtils._conjugate_be(count=1, tense='present')
        'is'
        >>> EnglishUtils._conjugate_be(count=2, tense='present-perfect')
        'have been'
        >>> EnglishUtils._conjugate_be(count=2, tense='would')
        'would be'

        :param count: The number of items in the subject
        :param tense: The verb tense (past, past-perfect, or present)
                      or a modal (can, could, may, might, must, shall, should, will would).

        """
        is_or_are = cls._BE_TENSES.get(tense, {}).get('singular' if count == 1 else 'plural')
        if is_or_are is None:
            if tense in cls._MODALS:
                is_or_are = f"{tense} be"
            else:
                raise ValueError(f"The tense given, {tense}, was neither a supported tense"
                                 f" ({cls.disjoined_list(sorted(list(cls._BE_TENSES.keys())))})"
                                 f" nor a modal ({cls.disjoined_list(sorted(list(cls._MODALS)))}).")
        return is_or_are  # possibly in some other tense. :)

    @classmethod
    def there_are(cls, items, *, kind: str = "thing", count: Optional[int] = None, there: str = "there",
                  capitalize=True, joiner=None, zero: object = "no", punctuate=None, punctuate_none=None,
                  use_article=False, show=True, context=None, tense='present', punctuation_mark: str = ".",
                  just_are=False, **joiner_options) -> str:
        """
        Constructs a sentence that enumerates a set of things.

        :param items: the items to enumerate
        :param kind: the kind of items being enumerated (default "thing")
        :param count: the number of items (defaults to the result of 'len(items)')
        :param there: the demonstrative or noun phrase that starts the sentence (default "there")
        :param capitalize: whether to capitalize the first letter of the sentence (default True)
        :param joiner: the joining function to join the items (default if None is just a commas-separated list)
        :param zero: the value to print instead of a numeric zero (default "no")
        :param punctuate: in the case of one or more values (not zero), whether to end with a period (default False)
        :param punctuate_none: in the case of no values or not showing values, whether to end with a period
               (default True if show is True, and otherwise is the same as the value of punctuate)
        :param punctuation_mark: if specified, something to use at the end if punctuating
        :param use_article: whether to put 'a' or 'an' in front of each option (default False)
        :param joiner_options: additional (keyword) options to be used with a joiner function if one is supplied
        :param show: whether to show the items if there are any (default True)
        :param context: an optional prepositional phrase indicating the context of the item(s) (default None)
        :param tense: one of 'past', 'present', 'future', 'conditional', or 'hypothetical' for the verbs used
        :param just_are: whether to stop at "There is" or "There are" without anything else.

        By far the most common uses are likely to be:

        >>> there_are(['Joe', 'Sally'], kind="user")
        "There are 2 users: Joe, Sally"
        >>> there_are(['Joe'], kind="user")
        "There is 1 user: Joe"
        >>> there_are([], kind="user")
        "There are no users."

        There are various control options. For example:

        >>> there_are(['Joe', 'Sally'], kind="user", joiner=conjoined_list, punctuate=True)
        "There are 2 users: Joe and Sally."
        >>> there_are(['Joe'], kind="user", joiner=conjoined_list, punctuate=True)
        "There is 1 user: Joe."
        >>> there_are([], kind="user", joiner=conjoined_list, punctuate=True)
        "There are no users."

        """

        if punctuate is None:
            punctuate = False if show else True

        if punctuate_none is None:
            punctuate_none = True if show else punctuate

        there = capitalize1(there) if capitalize else there
        n = len(items) if count is None else count
        # If the items is not in the tenses table, it's assumed to be a modal like 'might', 'may', 'must', 'can' etc.
        is_or_are = cls._conjugate_be(count=n, tense=tense)
        part0 = f"{there} {is_or_are}"
        if just_are:
            return part0
        part1 = f"{part0} {n_of(n, kind, num_format=lambda n, thing: zero if n == 0 else None)}"
        if context:
            part1 += f" {context}"
        if n == 0 or not show:
            punctuation = punctuation_mark if punctuate_none else ""
            return part1 + punctuation
        else:
            if use_article:
                items = [a_or_an(str(item)) for item in items]
            else:
                items = [str(item) for item in items]
            if joiner is None:
                joined = ", ".join(items)
            else:
                joined = joiner(items, **joiner_options)
            punctuation = punctuation_mark if punctuate else ""
            return f"{part1}: {joined}{punctuation}"

    @classmethod
    def _time_count_formatter(cls, n, unit):
        ignored(unit)
        if isinstance(n, float):
            return ("%.6f" % n).rstrip('0').rstrip('.')
        else:
            return n

    @classmethod
    def relative_time_string(cls, seconds, detailed=True):
        """
        Given a number of seconds, expresses that number of seconds in English.
        The seconds can be expressed either as a number or a datetime.timedelta.
        """
        result = []
        if isinstance(seconds, datetime.timedelta):
            seconds = seconds.total_seconds()
        remaining_seconds = seconds
        units_seen = False
        for unit_info in cls._TIME_UNITS:
            (unit_seconds, unit_name) = unit_info
            number_of_units = int(remaining_seconds // unit_seconds)
            remaining_seconds = remaining_seconds % unit_seconds
            if number_of_units != 0:
                units_seen = True
                result.append(cls.n_of(number_of_units, unit_name, num_format=cls._time_count_formatter))
            else:
                result.append(None)
        if not units_seen or remaining_seconds != 0:
            result.append(cls.n_of(remaining_seconds, "second", num_format=cls._time_count_formatter))
        if not detailed:
            abbreviated = []
            for item in result:
                if item or abbreviated:
                    stopping = bool(abbreviated)  # Stopping if this is the second item
                    abbreviated.append(item)
                    if stopping:
                        break
            result = abbreviated
        result = [item for item in result
                  if isinstance(item, str)]
        result = ", ".join(result)
        return result

    @classmethod
    def parse_relative_time_string(cls, s):
        parts = [x for x in s.split(' ') if x != '']
        if len(parts) % 2 != 0:
            raise ValueError(f"Relative time strings are an even number of tokens"
                             f" of the form '<n1> <unit1> <n2> <unit2>...': {s!r}")
        kwargs = {}
        for i in range(len(parts) // 2):
            # Canonicalize "1 week" or "1 weeks" to "weeks": 1.0 for inclusion as kwarg to timedelta
            # Uses specialized knowledge that all time units don't end in "s" but pluralize with "+s"
            value = float(parts[2 * i])
            units = parts[2 * i + 1].rstrip(',s') + "s"
            kwargs[units] = value
        try:
            return datetime.timedelta(**kwargs)
        except Exception:
            raise ValueError(f"Bad relative time string: {s!r}")

    @classmethod
    def disjoined_list(cls, items, conjunction: str = 'or', comma: Union[bool, str] = ",",
                       oxford_comma: Union[bool, str] = False, whitespace: str = " ",
                       nothing: Optional[str] = None) -> str:
        """
        Given a list of items, returns an English string that describes the option of any of them,
        joined by commas, as needed, and with the conjunction 'or' before the last item if there's more than one.

        For example:

        >>> disjoined_list(['something'])
        'something'
        >>> disjoined_list(['P', 'NP'])
        'P or NP'
        >>> disjoined_list(['this', 'that', 'the other'])
        'this, that or the other'
        >>> disjoined_list(['this', 'that', 'the other'], oxford_comma=True)
        'this, that, or the other'
        >>> disjoined_list(['this', 'that', 'the other'], comma=False)
        'this or that or the other'

        :param items: a list of items
        :param conjunction: a string (default 'or') to be used before the last item if there's more than one
        :param comma: a string (default ',') to use as a comma. Semicolon (';') is the most obvious other choice,
                      or False to indicate that the conjunction should be used between all elements.
        :param oxford_comma: a boolean (default False) saying whether to use a so-called 'Oxford comma',
                             or a string to use as that comma.
        :param whitespace: what to use as separator whitespace (default ' ')
        :param nothing: a string to use if there are no items, to avoid an error being raised.
        """

        return cls.conjoined_list(items, conjunction=conjunction, comma=comma, oxford_comma=oxford_comma,
                                  whitespace=whitespace, nothing=nothing)

    @classmethod
    def _item_strings(cls, items):
        return [str(x) for x in (sorted(items) if isinstance(items, set) else items)]

    @classmethod
    def conjoined_list(cls, items, conjunction: str = 'and', comma: Union[bool, str] = ",",
                       oxford_comma: Union[bool, str] = False, whitespace: str = " ",
                       nothing: Optional[str] = None) -> str:
        """
        Given a list of items, returns an English string that describes the collection of all of them,
        joined by commas, as needed, and with the conjunction 'and' before the last item if more than one item.

        For example:

        >>> conjoined_list(['something'])
        'something'
        >>> conjoined_list(['yin', 'yang'])
        'yin and yang'
        >>> conjoined_list(['up', 'down', 'all around'])
        'up, down and all around'
        >>> conjoined_list(['up', 'down', 'all around'], oxford_comma=True)
        'up, down, and all around'
        >>> conjoined_list(['up', 'down', 'all around'], comma=False)
        'up and down and all around'

        :param items: a list of items.
            If a dictionary is given, a list of its keys will be used.
            If a set is used, it will be converted to a sorted list.
        :param conjunction: a string (default 'and') to be used before the last item if there's more than one
        :param comma: a string (default ',') to use as a comma. Semicolon (';') is the most obvious other choice,
                      or False to indicate that the conjunction should be used between all elements.
        :param oxford_comma: a boolean (default False) saying whether to use an 'Oxford comma' (ask Google),
                             or a string to use as the Oxford comma.
        :param whitespace: what to use as separator whitespace (default ' ')
        :param nothing: a string to use if there are no items, to avoid an error being raised.
        """

        items = cls._item_strings(items)
        assert isinstance(conjunction, str), "The 'conjunction' argument must a string or boolean."
        conj = conjunction + whitespace

        if comma is False:
            sep = whitespace + conj
            oxford_comma = False  # It would be odd to
        elif comma is True:
            sep = "," + whitespace
        else:
            assert isinstance(comma, str), "The 'comma' argument must a string or boolean."
            sep = comma + whitespace

        if oxford_comma is False:
            final_sep = whitespace
        elif oxford_comma is True:
            final_sep = sep
        else:
            assert isinstance(oxford_comma, str), "The 'oxford_comma' argument must a string or boolean."
            final_sep = oxford_comma + whitespace

        if isinstance(items, dict):
            items = list(items.keys())
        elif isinstance(items, set):
            items = sorted(items)

        n = len(items)

        if n == 0:
            if nothing:
                return nothing
            else:
                raise ValueError("Cannot construct a conjoined list with no elements.")
        elif n == 1:
            return str(items[0])
        elif n == 2:
            return f"{items[0]}{whitespace}{conj}{items[1]}"
        else:
            return sep.join(items[:-1]) + f"{final_sep}{conj}{items[-1]}"


# Export specific useful functions

a_or_an = EnglishUtils.a_or_an

maybe_pluralize = EnglishUtils.maybe_pluralize

n_of = EnglishUtils.n_of

conjoined_list = EnglishUtils.conjoined_list
disjoined_list = EnglishUtils.disjoined_list

relative_time_string = EnglishUtils.relative_time_string
parse_relative_time_string = EnglishUtils.parse_relative_time_string

select_a_or_an = EnglishUtils.select_a_or_an

string_pluralize = EnglishUtils.string_pluralize

there_are = EnglishUtils.there_are

must_be_one_of = EnglishUtils.must_be_one_of
